FORMAT_VERSION = 131

NEW_FLOAT_EXT = 'F'      # 70  [Float64:IEEE float]
BIT_BINARY_EXT = 'M'     # 77  [UInt32:Len, UInt8:Bits, Len:Data]
SMALL_INTEGER_EXT = 'a'  # 97  [UInt8:Int]
INTEGER_EXT = 'b'        # 98  [Int32:Int]
FLOAT_EXT = 'c'          # 99  [31:Float String] Float in string format (formatted "%.20e", sscanf "%lf"). Superseded by NEW_FLOAT_EXT
ATOM_EXT = 'd'           # 100 [UInt16:Len, Len:AtomName] max Len is 255
REFERENCE_EXT = 'e'      # 101 [atom:Node, UInt32:ID, UInt8:Creation]
PORT_EXT = 'f'           # 102 [atom:Node, UInt32:ID, UInt8:Creation]
PID_EXT = 'g'            # 103 [atom:Node, UInt32:ID, UInt32:Serial, UInt8:Creation]
SMALL_TUPLE_EXT = 'h'    # 104 [UInt8:Arity, N:Elements]
LARGE_TUPLE_EXT = 'i'    # 105 [UInt32:Arity, N:Elements]
NIL_EXT = 'j'            # 106 empty list
STRING_EXT = 'k'         # 107 [UInt32:Len, Len:Characters]
LIST_EXT = 'l'           # 108 [UInt32:Len, Elements, Tail]
BINARY_EXT = 'm'         # 109 [UInt32:Len, Len:Data]
SMALL_BIG_EXT = 'n'      # 110 [UInt8:n, UInt8:Sign, n:nums]
LARGE_BIG_EXT = 'o'      # 111 [UInt32:n, UInt8:Sign, n:nums]
NEW_FUN_EXT = 'p'        # 112 [UInt32:Size, UInt8:Arity, 16*Uint6-MD5:Uniq, UInt32:Index, UInt32:NumFree, atom:Module, int:OldIndex, int:OldUniq, pid:Pid, NunFree*ext:FreeVars]
EXPORT_EXT = 'q'         # 113 [atom:Module, atom:Function, smallint:Arity]
NEW_REFERENCE_EXT = 'r'  # 114 [UInt16:Len, atom:Node, UInt8:Creation, Len*UInt32:ID]
SMALL_ATOM_EXT = 's'     # 115 [UInt8:Len, Len:AtomName]
MAP_EXT = 't'            # 116 [UInt32:Airty, N:Pairs]
FUN_EXT = 'u'            # 117 [UInt4:NumFree, pid:Pid, atom:Module, int:Index, int:Uniq, NumFree*ext:FreeVars]
COMPRESSED = 'P'         # 80  [UInt4:UncompressedSize, N:ZlibCompressedData]




"""
Types based on work from Samuel Stauffer's `python-erlastic` library. See COPYING.
"""

import six


__all__ = ['Atom', 'Reference', 'Port', 'PID', 'Export']


class Atom(six.text_type):
    def __repr__(self):
        return 'Atom(%s)' % super(Atom, self).__repr__()


class Reference(object):
    __slots__ = ['node', 'ref_id', 'creation']

    def __init__(self, node, ref_id, creation):
        if not isinstance(ref_id, tuple):
            ref_id = tuple(ref_id)

        self.node = node
        self.ref_id = ref_id
        self.creation = creation

    def __cmp__(self, other):
        if not isinstance(other, Reference):
            return 1
        if self.node == other.node and self.ref_id == other.ref_id and self.creation == other.creation:
            return 0
        return -1

    def __str__(self):
        return '#Ref<%d.%s>' % (self.creation, '.'.join(str(i) for i in self.ref_id))

    def __repr__(self):
        return '%s::%s' % (self, self.node)


class Port(object):
    __slots__ = ['node', 'port_id', 'creation']

    def __init__(self, node, port_id, creation):
        self.node = node
        self.port_id = port_id
        self.creation = creation

    def __cmp__(self, other):
        if not isinstance(other, Port):
            return 1
        if self.node == other.node and self.port_id == other.port_id and self.creation == other.creation:
            return 0
        return -1

    def __str__(self):
        return '#Port<%d.%d>' % (self.creation, self.port_id)

    def __repr__(self):
        return '%s::%s' % (self, self.node)


class PID(object):
    __slots__ = ['node', 'pid_id', 'serial', 'creation']

    def __init__(self, node, pid_id, serial, creation):
        self.node = node
        self.pid_id = pid_id
        self.serial = serial
        self.creation = creation

    def __cmp__(self, other):
        if not isinstance(other, PID):
            return 1
        if self.node == other.node and self.pid_id == other.pid_id and self.serial == other.serial and self.creation == other.creation:
            return 0
        return -1

    def __str__(self):
        return '<%d.%d.%d>' % (self.creation, self.pid_id, self.serial)

    def __repr__(self):
        return '%s::%s' % (self, self.node)


class Export(object):
    __slots__ = ['module', 'function', 'arity']

    def __init__(self, module, function, arity):
        self.module = module
        self.function = function
        self.arity = arity

    def __cmp__(self, other):
        if not isinstance(other, Export):
            return 1
        if self.module == other.module and self.function == other.function and self.arity == other.arity:
            return 0
        return -1

    def __str__(self):
        return '#Fun<%s.%s.%d>' % (self.module, self.function, self.arity)

    def __repr__(self):
        return str(self)



import ctypes

FORMAT_VERSION = 131
NEW_FLOAT_EXT = ord('F')       # 70  [Float64:IEEE float]
BIT_BINARY_EXT = ord('M')      # 77  [UInt32:Len, UInt8:Bits, Len:Data]
COMPRESSED = ord('P')          # 80  [UInt4:UncompressedSize, N:ZlibCompressedData]
SMALL_INTEGER_EXT = ord('a')   # 97  [UInt8:Int]
INTEGER_EXT = ord('b')         # 98  [Int32:Int]
FLOAT_EXT = ord('c')           # 99  [31:Float String] Float in string format (formatted "%.20e", sscanf "%lf"). Superseded by NEW_FLOAT_EXT
ATOM_EXT = ord('d')            # 100 [UInt16:Len, Len:AtomName] max Len is 255
REFERENCE_EXT = ord('e')       # 101 [atom:Node, UInt32:ID, UInt8:Creation]
PORT_EXT = ord('f')            # 102 [atom:Node, UInt32:ID, UInt8:Creation]
PID_EXT = ord('g')             # 103 [atom:Node, UInt32:ID, UInt32:Serial, UInt8:Creation]
SMALL_TUPLE_EXT = ord('h')     # 104 [UInt8:Arity, N:Elements]
LARGE_TUPLE_EXT = ord('i')     # 105 [UInt32:Arity, N:Elements]
NIL_EXT = ord('j')             # 106 empty list
STRING_EXT = ord('k')          # 107 [UInt16:Len, Len:Characters]
LIST_EXT = ord('l')            # 108 [UInt32:Len, Elements, Tail]
BINARY_EXT = ord('m')          # 109 [UInt32:Len, Len:Data]
SMALL_BIG_EXT = ord('n')       # 110 [UInt8:n, UInt8:Sign, n:nums]
LARGE_BIG_EXT = ord('o')       # 111 [UInt32:n, UInt8:Sign, n:nums]
NEW_FUN_EXT = ord('p')         # 112 [UInt32:Size, UInt8:Arity, 16*Uint6-MD5:Uniq, UInt32:Index, UInt32:NumFree, atom:Module, int:OldIndex, int:OldUniq, pid:Pid, NunFree*ext:FreeVars]
EXPORT_EXT = ord('q')          # 113 [atom:Module, atom:Function, smallint:Arity]
NEW_REFERENCE_EXT = ord('r')   # 114 [UInt16:Len, atom:Node, UInt8:Creation, Len*UInt32:ID]
SMALL_ATOM_EXT = ord('s')      # 115 [UInt8:Len, Len:AtomName]
MAP_EXT = ord('t')             # 116 [UInt32:Airty, N:Pairs]
FUN_EXT = ord('u')             # 117 [UInt4:NumFree, pid:Pid, atom:Module, int:Index, int:Uniq, NumFree*ext:FreeVars]
ATOM_UTF8_EXT = ord('v')       # 118 [UInt16:Len, Len:AtomName] max Len is 255 characters (up to 4 bytes per)
SMALL_ATOM_UTF8_EXT = ord('w') # 119 [UInt8:Len, Len:AtomName]


def erlpack_append_small_atom_ext(pk: bytearray, b: bytes):
	pk.extend([SMALL_ATOM_EXT, len(b)])
	pk.extend(b)
	return 0

def erlpack_append_nil(pk: bytearray)->int:
	return erlpack_append_small_atom_ext(pk, b"nil")
def erlpack_append_true(pk: bytearray)->int:
	return erlpack_append_small_atom_ext(pk, b"true")
def erlpack_append_false(pk: bytearray)->int:
	return erlpack_append_small_atom_ext(pk, b"false")

def erlpack_append_small_integer(pk: bytearray, d: int)->int:
	pk.extend([SMALL_INTEGER_EXT, d])
	return 0
class asUnsigned(ctypes.Union):_fields_=[
	("i32",ctypes.c_int32),
	("u32",ctypes.c_uint32)
]
def erlpack_append_integer(pk: bytearray, d: int)->int:
	pk.append(INTEGER_EXT)
	pun = asUnsigned()
	pun.i32 = d
	pk.extend(pun.u32.to_bytes(4, "big"))
	return 0
def erlpack_append_unsigned_long_long(pk: bytearray, d: int)->int:
	pk.append(SMALL_BIG_EXT)
	buf = []
	while (d > 0):
		buf.append(d & 0xFF)
		d >>= 8
	pk.append(len(buf))
	pk.append(0)
	pk.extend(buf)
	return 0
def erlpack_append_long_long(pk: bytearray, d: int)->int:
	pk.append(SMALL_BIG_EXT)
	ull = -d if d < 0 else d
	buf = []
	while (ull > 0):
		buf.append(ull & 0xFF)
		ull >>= 8
	pk.append(len(buf))
	pk.append(1 if d < 0 else 0)
	pk.extend(buf)
	return 0

class typePunner(ctypes.Union):_fields_=[
	("ui64",ctypes.c_uint64),
	("df",ctypes.c_double)
]
def erlpack_append_double(pk: bytearray, f: float)->int:
	pk.append(NEW_FLOAT_EXT)
	p =typePunner()
	p.df = f
	pk.extend(p.ui64.to_bytes(8, "big"))
	return 0

def erlpack_append_atom(pk: bytearray, b: bytes)->int:
	if (len(b) < 255): return erlpack_append_small_atom_ext(pk, b)

	pk.append(ATOM_EXT)
	if (len(b) > 0xFFFF): return 1
	
	pk.extend(len(b).to_bytes(2, "big"))
	pk.extend(b)
	return 0

def erlpack_append_binary(pk: bytearray, b: bytes)->int:
	pk.append(BINARY_EXT)
	pk.extend(len(b).to_bytes(4, "big"))
	pk.extend(b)
	return 0

def erlpack_append_tuple_header(pk: bytearray, length: int)->int:
	if (length < 256): pk.extend([SMALL_TUPLE_EXT, length])
	else:
		pk.append(LARGE_TUPLE_EXT)
		pk.extend(length.to_bytes(4, "big"))
	return 0

def erlpack_append_nil_ext(pk: bytearray)->int:
	pk.append(NIL_EXT)
	return 0

def erlpack_append_list_header(pk: bytearray, length: int)->int:
	pk.append(LIST_EXT)
	pk.extend(length.to_bytes(4, "big"))
	return 0

def erlpack_append_map_header(pk: bytearray, length: int)->int:
	pk.append(MAP_EXT)
	pk.extend(length.to_bytes(4, "big"))
	return 0

def erlpack_append_version(pk: bytearray)->int:
	pk.append(FORMAT_VERSION)
	return 0


def erlpack_append_string(pk: bytearray, b: bytes)->int:
	pk.append(STRING_EXT)
	pk.extend(len(b).to_bytes(2, "big"))
	pk.extend(b)
	return 0

#dummy as unused
erlpack_append_int=0
erlpack_append_atom_utf8=0



from erlpack.types import Atom

DEFAULT_RECURSE_LIMIT = 256
BIG_BUF_SIZE = 1024 * 1024 * 2
INITIAL_BUFFER_SIZE = 1024 * 1024
MAX_SIZE = (2 ** 32) - 1
NULL = 0

from .encoder import erlpack_append_int, erlpack_append_nil, erlpack_append_true, erlpack_append_false, erlpack_append_version, erlpack_append_small_integer, erlpack_append_integer, erlpack_append_unsigned_long_long, erlpack_append_long_long, erlpack_append_double, erlpack_append_atom, erlpack_append_atom_utf8, erlpack_append_binary, erlpack_append_string, erlpack_append_tuple_header, erlpack_append_nil_ext, erlpack_append_list_header, erlpack_append_map_header

class EncodingError(Exception):
    pass


class ErlangTermEncoder(object):
    # cdef erlpack_buffer pk
    # cdef char* _encoding
    # cdef char* _unicode_errors
    # cdef char* _unicode_type
    # cdef object _encode_hook
    # cdef bool _in_use

    def __cinit__(self):
        self.pk = b""

    def __init__(self, encoding=b'utf-8', unicode_errors=b'strict', unicode_type=b'binary', encode_hook=None):
        if encoding is None:
            self._encoding = NULL
            self._unicode_errors = NULL
        else:
            if isinstance(encoding, str):
                _encoding = encoding.encode('ascii')
            else:
                _encoding = encoding

            if isinstance(unicode_errors, str):
                _unicode_errors = unicode_errors.encode('ascii')
            else:
                _unicode_errors = unicode_errors

            self._encoding = _encoding.decode("utf-8")
            self._unicode_errors = _unicode_errors.decode("utf-8")

        self._unicode_type = unicode_type
        self._encode_hook = encode_hook
        self._in_use = False

    def _ensure_buf(self):
        """
        Ensures that a buffer is available to be written to when serializing data.

        If there is no buffer, allocate one sized to `INITIAL_BUFFER_SIZE`. If allocation
        fails, raise a MemoryError.
        """
        self.pk = bytearray()
        # if self.pk.buf != NULL:
        #     self.pk.length = 0

        # else:
        #     self.pk.buf = <char*> malloc(INITIAL_BUFFER_SIZE)
        #     if self.pk.buf == NULL:
        #         raise MemoryError('Unable to allocate buffer')

        #     self.pk.allocated_size = INITIAL_BUFFER_SIZE
        #     self.pk.length = 0

    def _free_big_buf(self):
        """
        If the buffer is larger than `BIG_BUF_SIZE`, free it, so that packing large data does not hold onto
        the big buffer after the serialization is complete.
        """
        if len(self.pk) >= BIG_BUF_SIZE:
            self.pk = bytearray()

    def __dealloc__(self):
        del self.pk

    def _pack(self, o, nest_limit=DEFAULT_RECURSE_LIMIT):
        # cdef int ret
        # cdef long long llval
        # cdef unsigned long long ullval
        # cdef long longval
        # cdef double doubleval
        # cdef size_t sizeval
        # cdef dict d
        # cdef object obj

        if nest_limit < 0:
            raise EncodingError('Exceeded recursion limit')

        if o is None:
            ret = erlpack_append_nil(self.pk)

        elif o is True:
            ret = erlpack_append_true(self.pk)

        elif o is False:
            ret = erlpack_append_false(self.pk)

        elif isinstance(o, int):
            if 0 <= o <= 255:
                ret = erlpack_append_small_integer(self.pk, o)

            elif -2147483648 <= o <= 2147483647:
                ret = erlpack_append_integer(self.pk, o)

            else:
                if o > 0:
                    # this assures a overflow error
                    o.to_bytes(8)
                    ret = erlpack_append_unsigned_long_long(self.pk, o)

                else:
                    # this assures a overflow error
                    o.to_bytes(8, signed=True)
                    ret = erlpack_append_long_long(self.pk, o)

        elif isinstance(o, float):
            ret = erlpack_append_double(self.pk, o)

        elif isinstance(o, Atom):
            # TODO: Erlang can support utf-8 atoms, but until all of the
            # clients we know can speak it, we are going to continue sending
            # the latin-1 encoded deprecated style.
            obj = o.encode('latin-1', 'strict')
            ret = erlpack_append_atom(self.pk, obj)

        elif isinstance(o, bytes):
            ret = erlpack_append_binary(self.pk, o)

        elif isinstance(o, str):
            ret = self._encode_unicode(o)

        elif isinstance(o, tuple):
            if len(o) > MAX_SIZE:
                raise ValueError('tuple is too large')

            ret = erlpack_append_tuple_header(self.pk, len(o))
            if ret != 0:
                return ret

            for item in o:
                ret = self._pack(item, nest_limit - 1)
                if ret != 0:
                    return ret

        elif isinstance(o, list):
            sizeval = len(o)
            if sizeval == 0:
                ret = erlpack_append_nil_ext(self.pk)
            else:

                if sizeval > MAX_SIZE:
                    raise ValueError("list is too large")

                ret = erlpack_append_list_header(self.pk, sizeval)
                if ret != 0:
                    return ret

                for item in o:
                    ret = self._pack(item, nest_limit - 1)
                    if ret != 0:
                        return ret

                ret = erlpack_append_nil_ext(self.pk)

        # elif PyDict_CheckExact(o): # TODO: check this is the same as the replacment
        elif isinstance(o, dict):
            sizeval = len(o)

            if sizeval > MAX_SIZE:
                raise ValueError("dict is too large")

            ret = erlpack_append_map_header(self.pk, sizeval)
            if ret != 0:
                return ret

            for k, v in o.items():
                ret = self._pack(k, nest_limit - 1)
                if ret != 0:
                    return ret

                ret = self._pack(v, nest_limit - 1)
                if ret != 0:
                    return ret

        elif hasattr(o, '__erlpack__'):
            obj = o.__erlpack__()
            return self._pack(obj, nest_limit - 1)

        else:
            if self._encode_hook:
                obj = self._encode_hook(o)
                if obj is not None:
                    return self._pack(obj, nest_limit - 1)

            raise NotImplementedError('Unable to serialize %r' % o)

        return ret

    def _encode_unicode(self, obj):
        if not self._encoding:
            return self._pack([ord(x) for x in obj])

        st = obj.encode(self._encoding, self._unicode_errors)
        size = len(st)

        if self._unicode_type == b'binary':
            if size > MAX_SIZE:
                raise ValueError('unicode string is too large using unicode type binary')

            return erlpack_append_binary(self.pk, st)

        elif self._unicode_type == b'str':
            if size > 0xFFF:
                raise ValueError('unicode string is too large using unicode type str')

            return erlpack_append_string(self.pk, st, size)

        else:
            raise TypeError('Unknown unicode encoding type %s' % self._unicode_type)

    def pack(self, obj):
        self._ensure_buf()
        if self._in_use:
            raise RuntimeError("Attempting to reuse an ErlangTermEncoder that is currently encoding something")

        self._in_use = True
        try:
            ret = erlpack_append_version(self.pk)
            if ret == -1:
                raise MemoryError

            ret = self._pack(obj, DEFAULT_RECURSE_LIMIT)
            if ret == -1:
                raise MemoryError
            elif ret:  # should not happen.
                raise TypeError('_pack returned code(%s)' % ret)

            return bytes(self.pk)
        finally:
            self._free_big_buf()
            self._in_use = False




"""
Unpacker based on work from Samuel Stauffer's `python-erlastic` library. See COPYING.
"""

from __future__ import division
import struct
import zlib
from .constants import *
from .types import *

__all__ = ['ErlangTermDecoder', 'ErlangTermDecodeError']


class ErlangTermDecodeError(Exception):
    pass


# noinspection PyMethodMayBeStatic,PyPep8Naming,PyShadowingBuiltins,PyUnusedLocal
class ErlangTermDecoder(object):
    # def object encoding
    def __init__(self, encoding=None):
        self.encoding = encoding

    def loads(self, bytes, offset=0):
        version = ord(bytes[offset:offset+1])
        if version != FORMAT_VERSION:
            raise ErlangTermDecodeError('Bad version number. Expected %d found %d' % (FORMAT_VERSION, version))
        return self.decode_part(bytes, offset + 1)[0]

    def decode_part(self, bytes, offset=0):
        opcode = bytes[offset:offset+1]

        if opcode == b'a':
            return self.decode_a(bytes, offset + 1)

        elif opcode == b'b':
            return self.decode_b(bytes, offset + 1)

        elif opcode == b'c':
            return self.decode_c(bytes, offset + 1)

        elif opcode == b'F':
            return self.decode_F(bytes, offset + 1)

        elif opcode == b'd':
            return self.decode_d(bytes, offset + 1)

        elif opcode == b's':
            return self.decode_s(bytes, offset + 1)

        elif opcode == b'v':
            return self.decode_v(bytes, offset + 1)

        elif opcode == b'w':
            return self.decode_w(bytes, offset + 1)

        elif opcode == b't':
            return self.decode_t(bytes, offset + 1)

        elif opcode == b'h':
            return self.decode_h(bytes, offset + 1)

        elif opcode == b'i':
            return self.decode_i(bytes, offset + 1)

        elif opcode == b'j':
            return self.decode_j(bytes, offset + 1)

        elif opcode == b'k':
            return self.decode_k(bytes, offset + 1)

        elif opcode == b'l':
            return self.decode_l(bytes, offset + 1)

        elif opcode == b'm':
            return self.decode_m(bytes, offset + 1)

        elif opcode == b'n':
            return self.decode_n(bytes, offset + 1)

        elif opcode == b'o':
            return self.decode_o(bytes, offset + 1)

        elif opcode == b't':
            return self.decode_t(bytes, offset + 1)

        elif opcode == b'e':
            return self.decode_e(bytes, offset + 1)

        elif opcode == b'r':
            return self.decode_r(bytes, offset + 1)

        elif opcode == b'f':
            return self.decode_f(bytes, offset + 1)

        elif opcode == b'g':
            return self.decode_g(bytes, offset + 1)

        elif opcode == b'q':
            return self.decode_q(bytes, offset + 1)

        elif opcode == b'P':
            return self.decode_P(bytes, offset + 1)

        else:
            raise ValueError('Unexpected opcode %s' % opcode)

    def decode_a(self, bytes, offset):
        """SMALL_INTEGER_EXT"""
        return ord(bytes[offset:offset+1]), offset + 1

    def decode_b(self, bytes, offset):
        """INTEGER_EXT"""
        return struct.unpack('>l', bytes[offset:offset + 4])[0], offset + 4

    def decode_c(self, bytes, offset):
        """FLOAT_EXT"""
        return float(bytes[offset:offset + 31].split('\x00', 1)[0]), offset + 31

    def decode_F(self, bytes, offset):
        """NEW_FLOAT_EXT"""
        return struct.unpack('>d', bytes[offset:offset + 8])[0], offset + 8

    def decode_d(self, bytes, offset):
        """ATOM_EXT"""
        atom_len, = struct.unpack('>H', bytes[offset:offset + 2])
        offset += 2
        atom = bytes[offset:offset + atom_len]
        return self.convert_atom(atom), offset + atom_len

    def decode_s(self, bytes, offset):
        """SMALL_ATOM_EXT"""
        atom_len = ord(bytes[offset:offset+1])
        offset += 1
        atom = bytes[offset:offset + atom_len]
        return self.convert_atom(atom), offset + atom_len

    def decode_v(self, bytes, offset):
        """ATOM_UTF8_EXT"""
        atom_len, = struct.unpack('>H', bytes[offset:offset + 2])
        offset += 2
        atom = bytes[offset:offset + atom_len]
        return self.convert_atom(atom, encoding='utf-8'), offset + atom_len

    def decode_w(self, bytes, offset):
        """SMALL_ATOM_UTF8_EXT"""
        atom_len = ord(bytes[offset:offset+1])
        offset += 1
        atom = bytes[offset:offset + atom_len]
        return self.convert_atom(atom, encoding='utf-8'), offset + atom_len

    def decode_t(self, bytes, offset):
        """MAP_EXT"""
        arity, = struct.unpack('>L', bytes[offset:offset + 4])
        offset += 4
        kv = {}
        for i in range(arity):
            key, offset = self.decode_part(bytes, offset)
            value, offset = self.decode_part(bytes, offset)
            kv[key] = value
        return kv, offset

    def decode_h(self, bytes, offset):
        """SMALL_TUPLE_EXT"""
        arity = ord(bytes[offset:offset+1])
        offset += 1
        items = []
        for i in range(arity):
            val, offset = self.decode_part(bytes, offset)
            items.append(val)
        return tuple(items), offset

    def decode_i(self, bytes, offset):
        """LARGE_TUPLE_EXT"""
        arity, = struct.unpack('>L', bytes[offset:offset + 4])
        offset += 4
        items = []
        for i in range(arity):
            val, offset = self.decode_part(bytes, offset)
            items.append(val)
        return tuple(items), offset

    def decode_j(self, bytes, offset):
        """NIL_EXT"""
        return [], offset

    def decode_k(self, bytes, offset):
        """STRING_EXT"""
        length, = struct.unpack('>H', bytes[offset:offset + 2])
        offset += 2
        st = bytes[offset:offset + length]
        byte_elements_are_ints = isinstance(b'a'[0], int)
        if byte_elements_are_ints:
            st = list(st)
        else:
            st = [ord(x) for x in st]
        return st, offset + length

    def decode_l(self, bytes, offset):
        """LIST_EXT"""
        length, = struct.unpack('>L', bytes[offset:offset + 4])
        offset += 4
        items = []
        for i in range(length):
            val, offset = self.decode_part(bytes, offset)
            items.append(val)
        tail, offset = self.decode_part(bytes, offset)
        # noinspection PySimplifyBooleanCheck
        if tail != []:
            # TODO: Not sure what to do with the tail
            raise NotImplementedError('Lists with non empty tails are not supported')
        return items, offset

    def decode_m(self, bytes, offset):
        """BINARY_EXT"""
        length, = struct.unpack('>L', bytes[offset:offset + 4])
        offset += 4
        rv = bytes[offset:offset + length]
        if self.encoding:
            rv = rv.decode(self.encoding)
        return rv, offset + length

    def decode_n(self, bytes, offset):
        """SMALL_BIG_EXT"""
        n = ord(bytes[offset:offset+1])
        offset += 1
        return self.decode_bigint(n, bytes, offset)

    def decode_o(self, bytes, offset):
        """LARGE_BIG_EXT"""
        n, = struct.unpack('>L', bytes[offset:offset + 4])
        offset += 4
        return self.decode_bigint(n, bytes, offset)

    def decode_bigint(self, n, bytes, offset):

        pos = 0

        if offset + 1 + n > len(bytes):
            raise OverflowError("Overflown! %s %s" % (offset + 1 + n, len(bytes)))

        sign = bytes[offset]
        offset += 1

        if sign == 0 and n <= 8:
            ull = 0
            for i in range(n):
                ull |= bytes[offset] << pos
                pos += 8
                offset += 1

            return ull, offset

        val = 0
        b = 1
        for i in range(n):
            val += bytes[offset] * b
            b <<= 8
            offset += 1
        if sign != 0:
            val = -val
        return val, offset

    def decode_e(self, bytes, offset):
        """REFERENCE_EXT"""
        node, offset = self.decode_part(bytes, offset)
        if not isinstance(node, Atom):
            raise ErlangTermDecodeError('Expected atom while parsing REFERENCE_EXT, found %r instead' % node)
        reference_id, creation = struct.unpack('>LB', bytes[offset: offset + 5])
        offset += 5
        return Reference(node, [reference_id], creation), offset

    def decode_r(self, bytes, offset):
        """NEW_REFERENCE_EXT"""
        id_len, = struct.unpack('>H', bytes[offset:offset + 2])
        offset += 2
        node, offset = self.decode_part(bytes, offset)
        if not isinstance(node, Atom):
            raise ErlangTermDecodeError('Expected atom while parsing NEW_REFERENCE_EXT, found %r instead' % node)
        creation = ord(bytes[offset])
        offset += 1
        reference_id = struct.unpack('>%dL' % id_len, bytes[offset:offset + (4 * id_len)])
        offset += (4 * id_len)
        return Reference(node, reference_id, creation), offset

    def decode_f(self, bytes, offset):
        """PORT_EXT"""
        node, offset = self.decode_part(bytes, offset)
        if not isinstance(node, Atom):
            raise ErlangTermDecodeError('Expected atom while parsing PORT_EXT, found %r instead' % node)
        port_id, creation = struct.unpack('>LB', bytes[offset:offset + 5])
        offset += 5
        return Port(node, port_id, creation), offset

    def decode_g(self, bytes, offset):
        """PID_EXT"""
        node, offset = self.decode_part(bytes, offset)
        if not isinstance(node, Atom):
            raise ErlangTermDecodeError('Expected atom while parsing PID_EXT, found %r instead' % node)
        pid_id, serial, creation = struct.unpack('>LLB', bytes[offset:offset + 9])
        offset += 9
        return PID(node, pid_id, serial, creation), offset

    def decode_q(self, bytes, offset):
        """EXPORT_EXT"""
        module, offset = self.decode_part(bytes, offset)
        if not isinstance(module, Atom):
            raise ErlangTermDecodeError('Expected atom while parsing EXPORT_EXT, found %r instead' % module)
        function, offset = self.decode_part(bytes, offset)
        if not isinstance(function, Atom):
            raise ErlangTermDecodeError('Expected atom while parsing EXPORT_EXT, found %r instead' % function)
        arity, offset = self.decode_part(bytes, offset)
        if not isinstance(arity, int):
            raise ErlangTermDecodeError('Expected integer while parsing EXPORT_EXT, found %r instead' % arity)
        return Export(module, function, arity), offset + 1

    def decode_P(self, bytes, offset):
        """Compressed term"""
        usize, = struct.unpack('>L', bytes[offset:offset + 4])
        offset += 4
        bytes = zlib.decompress(bytes[offset:offset + usize])
        return self.decode_part(bytes, 0)

    def convert_atom(self, atom, encoding='latin1'):
        """Convert an atom (bytes) into an appropriate Python object,
        using specified encoding. The default encoding is latin1, which
        is the old-style encoding."""
        if atom == b'true':
            return True
        elif atom == b'false':
            return False
        elif atom == b'nil':
            return None
        return Atom(atom.decode(encoding))



from ._packer import ErlangTermEncoder
from ._unpacker import ErlangTermDecoder
from .types import Atom, Export, PID, Port, Reference

encoder = ErlangTermEncoder()
pack = encoder.pack

decoder = ErlangTermDecoder()
unpack = decoder.loads

__all__ = ['pack', 'unpack', 'Atom', 'Export', 'PID', 'Port', 'Reference', 'ErlangTermEncoder']



from setuptools import setup

setup(
    name='pure_py_erlpack',
    version='1.0.0',
    author='Jake Heinz',
    author_email='jh@discordapp.com',
    url="http://github.com/discord/erlpack",
    description='A erlang term encoder for Python.',
    license='Apache 2.0',
    zip_safe=False,
    package_dir={'': 'py'},
    packages=['erlpack'],
    install_requires=['six~=1.15'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)
